//==================================================
//--------- Programmed by Alexey Grebennikov--------
//-------- alexey.grebennikov@bakerhughes.com-------
//==================================================
//
// Top module of FFT block
// This module implements FFT for 128 samples
// Input data is provided from A/D converter as 16-bit samples
//
`include "parms.v"

module fft_top    (
  clk_i, rst_i, data_clk_i, re_data_i, re2_data_i, im_data_i, im2_data_i, conv_start_i, dvalid_i,
  rd_i,
  re_data_o, im_data_o, conv_rdy_o, adr_o, idx_o, test1_o, test2_o
  );
  input clk_i;                  // Input clock
  input rst_i;                  // Input reset signal
  input data_clk_i;             // Clock to shift input data
  input [13:0] re_data_i;       // Real part of input data 
  input [13:0] re2_data_i;      // Second input for real part of data
  input [13:0] im_data_i;       // Imaginary part of input data
  input [13:0] im2_data_i;      // Second input for imaginary part of data
  input conv_start_i;           // Signal to start conversion
  input dvalid_i;               // Data valid - signal to clock input data
  input rd_i;                   // Input signal to read one sample of conversion
  output [17:0] re_data_o;     // Output real part of conversion
  output [17:0] im_data_o;     // Output imaginary part of conversion
  output conv_rdy_o;            // Conversion completed
  output [6:0] adr_o;           // Address for output samples
  output [5:0] idx_o;           // Index of data set to be read from wb_fft module
  output [3:0] test1_o;         // First test output
  output [3:0] test2_o;         // Second test output
  
  reg conv_rdy_o;
  reg [6:0] adr_o;
  reg [17:0] re_data_o;
  reg [17:0] im_data_o;
  // Internal wires and registers    
  wire [17:0] twr_w;
  wire [17:0] twi_w;
  reg [5:0] adr_r;
  reg [7:0] cnt1;
  reg start_r;
  
  reg [17:0] re1_int [0:63];        // Arrays to store intermediate results of operations
  reg [17:0] im1_int [0:63];
  reg [17:0] re2_int [0:63];
  reg [17:0] im2_int [0:63];
  reg [17:0] re1_int_r;
  reg [17:0] im1_int_r;
  reg [17:0] re2_int_r;
  reg [17:0] im2_int_r;
  reg [17:0] re1_out [0:63];      // Final data arrays
  reg [17:0] im1_out [0:63];
  reg [17:0] re2_out [0:63];
  reg [17:0] im2_out [0:63];
  
  reg next_state1;
  reg next_state2;
  
  wire [17:0] re1_w;
  wire [17:0] im1_w;
  wire [17:0] re2_w;
  wire [17:0] im2_w;
  wire ready_w;

  reg [3:0] fft_fsm_state;
  reg [3:0] fft_fsm_next_state;
  reg fft_idle;
  reg stage1;
  reg stage2;
  reg stage3;
  reg stage4;
  reg stage5;
  reg stage6;
  reg stage7;
  reg [5:0] idx1;
  reg [5:0] idx2;
  reg [5:0] idx10;         // Second set of indexes
  reg [5:0] idx20;         // Second set of indexes
  reg buf2;                // When set data is recorded to the second buffer
  reg buf20;               // Second set of second data buffer flag
  reg [5:0] buf_cnt;       // Counter to set either first or second data buffers in stages 2 and higher
  reg stage2_twd;          // Flag to select twiddle factor for the second stage
  reg [1:0] stage3_twd;    // Flag to select twiddle factor for the third stage
  wire [3:0] mult_t1_w;
  wire first_w;
  wire idle_w;

  assign test1_o = fft_fsm_state;
  assign test2_o = mult_t1_w;
  wire [17:0] re_data_sx = {{4{re_data_i[13]}},re_data_i};        // Sign extended to 18-bit
  wire [17:0] re2_data_sx = {{4{re2_data_i[13]}},re2_data_i};
  wire [17:0] im_data_sx = {{4{im_data_i[13]}},im_data_i};
  wire [17:0] im2_data_sx = {{4{im2_data_i[13]}},im2_data_i};
  
  wire [17:0] re_data_w = (stage1) ? re_data_sx : re1_int_r ;
  wire [17:0] re2_data_w = (stage1) ? re2_data_sx : re2_int_r;
  wire [17:0] im_data_w = (stage1) ? im_data_sx : im1_int_r;
  wire [17:0] im2_data_w = (stage1) ? im2_data_sx : im2_int_r;
  
  reg [6:0] idx_r;        // Internal index register, MSB is overflow sign
  reg s_stage;            // Indicator of the seventh stage
  assign idx_o = idx_r[5:0];

  
    //-----------------------Chipscope logic-----------------------------
    wire [35:0] scope_cntr_w;

    chipscope_icon_v1_06_a_0 scope_cntr1 (
      .CONTROL0(scope_cntr_w) // INOUT BUS [35:0]
  );
  
    chipscope_ila_v1_05_a_0 scope_ila1 (    // Sample on rising edge
      .CONTROL(scope_cntr_w), // INOUT BUS [35:0]
      .CLK(data_clk_i), // IN
      .TRIG0(test1_o), // IN BUS [3:0]
      .TRIG1({2'b00,re1_w}), // IN BUS [19:0]
      .TRIG2({2'b00,im1_w}), // IN BUS [19:0]
      .TRIG3({2'b00,re2_w}), // IN BUS [19:0]
      .TRIG4({2'b00,im2_w}), // IN BUS [19:0]
      .TRIG5({8'h77,2'b00,re_data_w}),            // 28
      .TRIG6({2'b00,im_data_w}),            // 20
      .TRIG7({2'b00,re2_data_w}),            // 20
      .TRIG8({2'b00,im2_data_w}),            // 20
      .TRIG9({2'b00,twr_w}),            // 20
      .TRIG10({2'b00,twi_w}),           // 20
      .TRIG11(mult_t1_w)               // IN BUS [3:0]
  );
  //----------------------End of chipscope logic-----------------------  

  
  // Modules instantiation
  
  mult mult_fft (
    .clk_i                        (data_clk_i),
    .rst_i                        (rst_i),
    .re1_i                        (re_data_w),
    .re2_i                        (re2_data_w),
    .im1_i                        (im_data_w),
    .im2_i                        (im2_data_w),
    .twr_i                        (twr_w),
    .twi_i                        (twi_w),
    .start_i                      (start_r),
    .test1_i                      (fft_fsm_state),
    .re1_o                        (re1_w),
    .im1_o                        (im1_w),
    .re2_o                        (re2_w),
    .im2_o                        (im2_w),
    .ready_o                      (ready_w),
    .first_o                      (first_w),
    .idle_o                       (idle_w),
    .test1_o                      (mult_t1_w)
    );

  trom trom_fft (
    .clk_i                        (data_clk_i),
    .rst_i                        (rst_i),
    .adr_i                        (adr_r),
    .twr_o                        (twr_w),
    .twi_o                        (twi_w)
    );
    
//---------------FSM--------------------------
  
  always @ (fft_fsm_state or conv_start_i or next_state1 or next_state2)      // Signals in the sensitivity list change on the posedge
    begin
      fft_fsm_next_state = `FFT_IDLE;
      fft_idle = 1'b0;
      stage1 = 1'b0;
      stage2 = 1'b0;
      stage3 = 1'b0;
      stage4 = 1'b0;
      stage5 = 1'b0;
      stage6 = 1'b0;
      stage7 = 1'b0;
      conv_rdy_o = 1'b0;
      case (fft_fsm_state)          // Synthesis full case
        `FFT_IDLE:
          begin
            fft_idle = 1'b1;
            if (conv_start_i)
              fft_fsm_next_state = `FFT_STAGE1;
            else
              fft_fsm_next_state = `FFT_IDLE;
          end
        `FFT_STAGE1:
          begin
            stage1 = 1'b1;
            if (next_state1 && ~next_state2)
              fft_fsm_next_state = `FFT_STAGE2;
            else
              fft_fsm_next_state = `FFT_STAGE1;
          end
        `FFT_STAGE2:
          begin
            stage2 = 1'b1;
            if (~next_state1 && next_state2)
              fft_fsm_next_state = `FFT_STAGE3;
            else
              fft_fsm_next_state = `FFT_STAGE2;
          end
        `FFT_STAGE3:
          begin
            stage3 = 1'b1;
            if (next_state1 && ~next_state2)
              fft_fsm_next_state = `FFT_STAGE4;
            else
              fft_fsm_next_state = `FFT_STAGE3;
          end
        `FFT_STAGE4:
          begin
            stage4 = 1'b1;
            if (~next_state1 && next_state2)
              fft_fsm_next_state = `FFT_STAGE5;
            else
              fft_fsm_next_state = `FFT_STAGE4;
          end
        `FFT_STAGE5:
          begin
            stage5 = 1'b1;
            if (next_state1 && ~next_state2)
              fft_fsm_next_state = `FFT_STAGE6;
            else
              fft_fsm_next_state = `FFT_STAGE5;
          end
        `FFT_STAGE6:
          begin
            stage6 = 1'b1;
            if (~next_state1 && next_state2)
              fft_fsm_next_state = `FFT_STAGE7;
            else
              fft_fsm_next_state = `FFT_STAGE6;
          end
        `FFT_STAGE7:
          begin
            stage7 = 1'b1;
            if (next_state1 && ~next_state2)
              fft_fsm_next_state = `FFT_DONE;
            else
              fft_fsm_next_state = `FFT_STAGE7;
          end
        `FFT_DONE:
          begin
            conv_rdy_o = 1'b1;
            fft_fsm_next_state = `FFT_IDLE;
          end
      endcase
    end
    
  always @ (posedge clk_i)
    begin
      if (!rst_i)
        begin
          fft_fsm_state <= `FFT_IDLE;
        end
      else
        begin
          fft_fsm_state <= fft_fsm_next_state;
        end
    end  

//--------------End of FSM--------------------
//
  always @ (negedge clk_i)
    begin
      if (!rst_i)
        begin
          adr_o <= 7'h0;  
        end
      else
        begin
          if (conv_rdy_o)
            adr_o <= 7'h0;
          else
            begin
              if (rd_i)
                begin
                  adr_o <= adr_o + 1'b1;
                end 
            end
            
        if (~adr_o[6])
          begin
            re_data_o <= re1_int[adr_o[5:0]];
            im_data_o <= im1_int[adr_o[5:0]]; 
          end
        else
          begin
            re_data_o <= re2_int[adr_o[5:0]];
            im_data_o <= im2_int[adr_o[5:0]];
          end    
        
        end
    end
    
  always @ (posedge data_clk_i)
    begin
      if (!rst_i)
        begin
          cnt1 <= 8'h0;
          idx_r <= 7'h0;
          start_r <= 1'b0;
          next_state1 <= 1'b0;
          next_state2 <= 1'b0;
          idx1 <= 6'h0;
          idx2 <= 6'h1;
          buf2 <= 1'b0;
        end
      else
        begin
          if (stage2 | stage4 | stage6)
            begin
              re1_int_r <= re1_int[idx_o];
              re2_int_r <= re2_int[idx_o];
              im1_int_r <= im1_int[idx_o];
              im2_int_r <= im2_int[idx_o];
            end
          else if (stage3 | stage5 | stage7)
            begin
              re1_int_r <= re1_out[idx_o];
              re2_int_r <= re2_out[idx_o];
              im1_int_r <= im1_out[idx_o];
              im2_int_r <= im2_out[idx_o];
            end
          if (stage1)
            begin
              idx10 <= 6'h0;
              idx20 <= 6'h2;
              buf20 <= 1'b0;
              buf_cnt <= 6'h0;
              if (buf2)
                begin
                  if (ready_w)
                    begin
                      re2_int[idx1] <= re1_w;
                      im2_int[idx1] <= im1_w;
                      re2_int[idx2] <= re2_w;
                      im2_int[idx2] <= im2_w;
                      idx1 <= idx1 + 2;
                      idx2 <= idx2 + 2;
                      buf2 <= 1'b0;
                    end
                end
              else        // Data to be recorded to the first buffer
                begin
                  if (ready_w)
                    begin
                      re1_int[idx2] <= re2_w;
                      im1_int[idx2] <= im2_w;
                      re1_int[idx1] <= re1_w;
                      im1_int[idx1] <= im1_w;
                      buf2 <= 1'b1;
                    end
                end
            end
          else
            begin
              if (buf20)
                begin
                  if (ready_w)
                    begin
                      if (stage2 | stage4 | stage6)
                        begin
                          re2_out[idx20] <= re2_w;
                          im2_out[idx20] <= im2_w;
                          re2_out[idx10] <= re1_w;
                          im2_out[idx10] <= im1_w;
                        end
                      else if (stage3 | stage5)
                        begin
                          re2_int[idx20] <= re2_w;
                          im2_int[idx20] <= im2_w;
                          re2_int[idx10] <= re1_w;
                          im2_int[idx10] <= im1_w;
                        end
                      if (stage2)
                        begin
                          if (idx_r == 7'b0111111)
                            begin
                              idx10 <= 6'h0;
                              idx20 <= 6'h4;
                              buf20 <= 1'b0;
                            end
                        else if (buf_cnt == 6'h1)
                          begin
                            buf20 <= 1'b0;
                            idx10 <= idx10 + 3;
                            idx20 <= idx20 + 3;
                          end
                        else
                          begin
                            idx10 <= idx10 + 1'b1;
                            idx20 <= idx20 + 1'b1;
                          end
                        if (buf_cnt != 6'h1)
                          buf_cnt <= buf_cnt + 1'b1;
                        else
                          buf_cnt <= 6'h0;
                        end
                      else if (stage3)
                        begin
                          if (idx_r == 7'b0111111)
                            begin
                              idx10 <= 6'h0;
                              idx20 <= 6'h8;
                              buf20 <= 1'b0;
                            end
                          else if (buf_cnt == 6'h3)
                            begin
                              buf20 <= 1'b0;
                              idx10 <= idx10 + 5;
                              idx20 <= idx20 + 5;
                            end
                          else
                            begin
                              idx10 <= idx10 + 1'b1;
                              idx20 <= idx20 + 1'b1;
                            end
                          if (buf_cnt != 6'h3)
                            buf_cnt <= buf_cnt + 1'b1;
                          else
                            buf_cnt <= 6'h0;
                        end
                      else if (stage4)
                        begin
                          if (idx_r == 7'b0111111)
                            begin
                              idx10 <= 6'h0;
                              idx20 <= 6'h10;
                              buf20 <= 1'b0;
                            end
                          else if (buf_cnt == 6'h7)
                            begin
                              buf20 <= 1'b0;
                              idx10 <= idx10 + 9;
                              idx20 <= idx20 + 9;
                            end
                          else
                            begin
                              idx10 <= idx10 + 1'b1;
                              idx20 <= idx20 + 1'b1;
                            end
                          if (buf_cnt != 6'h07)
                            buf_cnt <= buf_cnt + 1'b1;
                          else
                            buf_cnt <= 6'h0;
                        end
                      else if (stage5)
                        begin
                          if (idx_r == 7'b0111111)
                            begin
                              idx10 <= 6'h0;
                              idx20 <= 6'h20;
                              buf20 <= 1'b0;
                              s_stage <= 1'b0;
                            end
                          else if (buf_cnt == 6'h0F)
                            begin
                              buf20 <= 1'b0;
                              idx10 <= idx10 + 17;
                              idx20 <= idx20 + 17;
                            end
                          else
                            begin
                              idx10 <= idx10 + 1'b1;
                              idx20 <= idx20 + 1'b1;
                            end
                          if (buf_cnt != 6'h0F)
                            buf_cnt <= buf_cnt + 1'b1;
                          else
                            buf_cnt <= 6'h0;
                        end
                      else if (stage6)
                        begin
                          if (idx_r == 7'b0111111)
                            begin
                              idx10 <= 6'h0;
                              idx20 <= 6'h0;     
                              buf20 <= 1'b0;
                              s_stage <= 1'b1;
                            end
                          else if (buf_cnt == 6'h1F)
                            begin
                              buf20 <= 1'b0;
                              idx10 <= idx10 + 33;
                              idx20 <= idx20 + 33;
                            end
                          else
                            begin
                              idx10 <= idx10 + 1'b1;
                              idx20 <= idx20 + 1'b1;
                            end
                          if (buf_cnt != 6'h1F)
                            buf_cnt <= buf_cnt + 1'b1;
                          else
                            buf_cnt <= 6'h0;
                        end
/*                      else if (stage7)
                        begin
                        
                        end        */
                    end
                end
              else if (~stage7)
                begin
                  if (ready_w)
                    begin
                      if (stage2 | stage4 | stage6)
                        begin
                          re1_out[idx20] <= re2_w;
                          im1_out[idx20] <= im2_w;
                          re1_out[idx10] <= re1_w;
                          im1_out[idx10] <= im1_w;
                        end
                      else if (stage3 | stage5)
                        begin
                          re1_int[idx20] <= re2_w;
                          im1_int[idx20] <= im2_w;
                          re1_int[idx10] <= re1_w;
                          im1_int[idx10] <= im1_w;
                        end
                      if (stage2)
                        begin
                          if (buf_cnt == 6'h1)
                            begin
                              buf_cnt <= 6'h0;
                              buf20 <= 1'b1;
                              idx10 <= idx10 - 1;
                              idx20 <= idx20 - 1;
                            end
                          else
                            begin
                              idx10 <= idx10 + 1'b1;
                              idx20 <= idx20 + 1'b1;
                              buf_cnt <= buf_cnt + 1'b1;
                            end
                        end
                      else if (stage3)
                        begin
                          if (buf_cnt == 6'h3)
                            begin
                              buf_cnt <= 6'h0;
                              buf20 <= 1'b1;
                              idx10 <= idx10 - 3;
                              idx20 <= idx20 - 3;
                            end
                          else
                            begin
                              idx10 <= idx10 + 1'b1;
                              idx20 <= idx20 + 1'b1;
                              buf_cnt <= buf_cnt + 1'b1;
                            end
                        end
                      else if (stage4)
                        begin
                          if (buf_cnt == 6'h7)
                            begin
                              buf_cnt <= 6'h0;
                              buf20 <= 1'b1;
                              idx10 <= idx10 - 7;
                              idx20 <= idx20 - 7;
                            end
                          else
                            begin
                              idx10 <= idx10 + 1'b1;
                              idx20 <= idx20 + 1'b1;
                              buf_cnt <= buf_cnt + 1'b1;
                            end
                        end
                      else if (stage5)
                        begin
                          if (buf_cnt == 6'hF)
                            begin
                              buf_cnt <= 6'h0;
                              buf20 <= 1'b1;
                              idx10 <= idx10 - 15;
                              idx20 <= idx20 - 15;
                            end
                          else
                            begin
                              idx10 <= idx10 + 1'b1;
                              idx20 <= idx20 + 1'b1;
                              buf_cnt <= buf_cnt + 1'b1;
                            end
                        end
                      else if (stage6)
                        begin
                          if (buf_cnt == 6'h1F)
                            begin
                              buf_cnt <= 6'h0;
                              buf20 <= 1'b1;
                              idx10 <= idx10 - 31;
                              idx20 <= idx20 - 31;
                            end
                          else
                            begin
                              idx10 <= idx10 + 1'b1;
                              idx20 <= idx20 + 1'b1;
                              buf_cnt <= buf_cnt + 1'b1;
                            end
                        end
                    end
                end
              else if (stage7)
                begin
                  if (ready_w)
                    begin
                      re1_int[idx10] <= re1_w;
                      im1_int[idx10] <= im1_w;
                      re2_int[idx10] <= re2_w;
                      re2_int[idx10] <= im2_w;
                      idx10 <= idx10 + 1'b1;
                      if (idx_r == 7'b0111111)
                        begin
                          idx20 <= 6'h10;   // Used as indicator of 7'th stage completion 
                        end
                    end
                end
            end
          
          if (ready_w)
            begin
              if (idx_r == 7'b1000000)
                idx_r <= 7'b1;
              else
                idx_r <= idx_r + 1'b1;
            end
          else if (fft_idle)
            idx_r <= 7'h0;
            
          if (stage1)
            begin                
              adr_r <= 6'h0;        // Zero twiddle is used for all operations              
              if (idx_r == 7'b1000000)
                begin
                  next_state1 <= 1'b1;
                  next_state2 <= 1'b0;
                  stage2_twd <= 1'b0;
                end
            end
          else if (stage2)
            begin
              if (ready_w)
                stage2_twd <= ~stage2_twd;
              if (stage2_twd)
                adr_r <= 6'h20;     // 32 in decimal
              else
                adr_r <= 6'h0;
              if ((idx_r == 7'b1000000) && idx20[2])
                begin
                 next_state1 <= 1'b0;
                 next_state2 <= 1'b1;
                 stage3_twd <= 2'b00;
                end
            end
          else if (stage3)
            begin
              if (ready_w)
                stage3_twd <= stage3_twd + 1'b1;
              case (stage3_twd)     // Twiddle factors for this stage are 0,16,32,48
                2'b00: adr_r <= 6'h0;
                2'b01: adr_r <= 6'h10;
                2'b10: adr_r <= 6'h20;
                2'b11: adr_r <= 6'h30;
              endcase
                  
              if ((idx_r == 7'b1000000) && idx20[3])
                begin
                  next_state1 <= 1'b1;
                  next_state2 <= 1'b0;
                end
            end
          else if (stage4)
            begin
              if (ready_w)
                begin
                  adr_r <= adr_r + 8;
                // Twiddle factors for this stage are 0,8,16,24,32,40,48,56
                end
              
              if ((idx_r == 7'b1000000) && idx20[4])
                begin
                  next_state1 <= 1'b0;
                  next_state2 <= 1'b1;
                end
            end
          else if (stage5)
            begin
              if (ready_w)
                adr_r <= adr_r + 4;
                 // Twiddle factors for this stage are 0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60
                 // in hex:                            0, 4, 8, C,  10, 14, 18, 1C, 20, 24, 28, 2C, 30, 34, 38, 3C
             if ((idx_r == 7'b1000000) && idx20[5])
                begin
                  next_state1 <= 1'b1;
                  next_state2 <= 1'b0;
                end
            end
          else if (stage6)
            begin
              if (ready_w)
                adr_r <= adr_r + 2;   // Twiddle factor is incremented by 2
              if ((idx_r == 7'b1000000) && s_stage)
                begin
                  next_state1 <= 1'b0;
                  next_state2 <= 1'b1;
                end
            end
          else if (stage7)
            begin
              if (ready_w)
                adr_r <= adr_r + 1;   // The whole array of twiddle factors is used
              if ((idx_r == 7'b1000000) && idx20[4])
                begin
                  next_state1 <= 1'b1;
                  next_state2 <= 1'b0;
                end 
            end
          else if (conv_rdy_o)
            begin
              next_state1 <= 1'b0;
              next_state2 <= 1'b0;
            end
            
                         
          if (~fft_fsm_state[3])  // Fourth bit is one for FFT_DONE and FFT_IDLE states only
            begin               
              if (ready_w)
                cnt1 <= 8'h0;
              else
                cnt1 <= cnt1 + 1'b1;
              if (cnt1 == 8'h0)
                start_r <= 1'b1;
              else
                start_r <= 1'b0;
            end
          else
            begin
              start_r <= 1'b0;
              cnt1 <= 8'h0;
            end
                        
        end
        
    end
endmodule
